@@ -4,12 +4,16 @@ import logging
 from functools import lru_cache
 from copy import copy, deepcopy
 from collections import defaultdict
-from typing import Mapping, Tuple
+from typing import Mapping, Tuple, Type
 from linkml_runtime.utils.namespaces import Namespaces
-from deprecated.classic import deprecated
-
 from linkml_runtime.utils.context_utils import parse_import_map
 from linkml_runtime.linkml_model.meta import *
+from deprecated.classic import deprecated
+from linkml_runtime.utils.context_utils import parse_import_map
+from linkml_runtime.linkml_model.meta import *
+
+logger = logging.getLogger(__name__)
+
 
 MAPPING_TYPE = str  ## e.g. broad, exact, related, ...
 CACHE_SIZE = 1024
@@ -65,6 +69,7 @@ class SchemaUsage():
     slot: SlotDefinitionName
     metaslot: SlotDefinitionName
     used: ElementName
+    inferred: bool = None
 
 
 @dataclass
@@ -315,15 +320,29 @@ class SchemaView(object):
     @lru_cache()
     def slot_name_mappings(self) -> Dict[str, SlotDefinition]:
         """
-        Mapping between processed safe slot names and slots.
+        Mapping between processed safe slot names (following naming conventions)  and slots.
 
         For example, a slot may have name 'lives at', the code-safe version is `lives_at`
+
         :return: mapping from safe names to slot
         """
         m = {}
         for s in self.all_slots().values():
             m[underscore(s.name)] = s
+        return m
 
+    @lru_cache()
+    def class_name_mappings(self) -> Dict[str, ClassDefinition]:
+        """
+        Mapping between processed safe class names (following naming conventions) and classes.
+
+        For example, a class may have name 'named thing', the code-safe version is `NamedThing`
+
+        :return: mapping from safe names to class
+        """
+        m = {}
+        for s in self.all_classes().values():
+            m[camelcase(s.name)] = s
         return m
 
 
@@ -608,7 +627,6 @@ class SchemaView(object):
             e = self.get_subset(element, imports=imports)
         return e
 
-
     def get_uri(self, element: Union[ElementName, Element], imports=True, expand=False, native=False) -> str:
         """
         Return the CURIE or URI for a schema element. If the schema defines a specific URI, this is
@@ -663,6 +681,40 @@ class SchemaView(object):
                     return ns[pfx] + local_id
         return uri
 
+    @lru_cache(CACHE_SIZE)
+    def get_elements_applicable_by_identifier(self, identifier: str) -> List[str]:
+        """
+        Get a model element by identifier.  The model element corresponding to the given identifier as available via
+        the id_prefixes mapped to that element.
+
+        :param identifier:
+        :return: Optional[str]
+
+        """
+        elements = self.get_elements_applicable_by_prefix(self.namespaces().prefix_for(identifier))
+        if len(elements) == 0:
+            logger.warning("no element found for the given curie using id_prefixes attribute"
+                           ": %s, try get_mappings method?", identifier)
+        return elements
+
+    @lru_cache(CACHE_SIZE)
+    def get_elements_applicable_by_prefix(self, prefix: str) -> List[str]:
+        """
+        Get a model element by prefix. The model element corresponding to the given prefix as available via
+        the id_prefixes mapped to that element.
+
+        :param prefix: the prefix of a CURIE
+        :return: Optional[str]
+
+        """
+        applicable_elements = []
+        elements = self.all_element()
+        for category, category_element in elements.items():
+            if hasattr(category_element, 'id_prefixes') and prefix in category_element.id_prefixes:
+                applicable_elements.append(category_element.name)
+
+        return applicable_elements
+
     @lru_cache()
     def get_mappings(self, element_name: ElementName = None, imports=True, expand=False) -> Dict[MAPPING_TYPE, List[URIorCURIE]]:
         """
@@ -674,22 +726,67 @@ class SchemaView(object):
         :return: index keyed by mapping type
         """
         e = self.get_element(element_name, imports=imports)
-        m_dict = {
-            'self': [self.get_uri(element_name, imports=imports, expand=False)],
-            'native': [self.get_uri(element_name, imports=imports, expand=False, native=True)],
-            'exact': e.exact_mappings,
-            'narrow': e.narrow_mappings,
-            'broad': e.broad_mappings,
-            'related': e.related_mappings,
-            'close': e.close_mappings,
-            'undefined': e.mappings
-        }
+        if isinstance(e, ClassDefinition) or isinstance(e, SlotDefinition) or isinstance(e, TypeDefinition):
+            m_dict = {
+                'self': [self.get_uri(element_name, imports=imports, expand=False)],
+                'native': [self.get_uri(element_name, imports=imports, expand=False, native=True)],
+                'exact': e.exact_mappings,
+                'narrow': e.narrow_mappings,
+                'broad': e.broad_mappings,
+                'related': e.related_mappings,
+                'close': e.close_mappings,
+                'undefined': e.mappings
+            }
+        else:
+            m_dict = {}
         if expand:
             for k, vs in m_dict.items():
                 m_dict[k] = [self.expand_curie(v) for v in vs]
 
         return m_dict
 
+    @lru_cache()
+    def is_mixin(self, element_name: Union[ElementName, Element]):
+        """
+        Determines whether the given name is the name of a mixin
+        in the model. An element is a mixin if one of its properties is "is_mixin:true"
+
+        :param element_name: The name or alias of an element in the model
+        :return: boolean
+        """
+
+        element = self.get_element(element_name)
+        is_mixin = element.mixin if isinstance(element, Definition) else False
+        return is_mixin
+
+    @lru_cache()
+    def inverse(self, slot_name: SlotDefinition):
+        """
+        Determines whether the given name is a relationship, and if that relationship has an inverse, returns
+        the inverse.
+
+        :param slot_name: The name or alias of an element in the model
+        :return: inverse_name
+
+        """
+        element = self.get_element(slot_name)
+        inverse = element.inverse if isinstance(element, SlotDefinition) else False
+        if not inverse:
+            for inv_slot_name, slot_definition in self.all_slots().items():
+                if slot_definition.inverse == element.name:
+                    inverse = slot_definition.name
+        return inverse
+
+    def get_element_by_mapping(self, mapping_id: URIorCURIE) -> List[str]:
+        model_elements = []
+        elements = self.all_elements()
+        for el in elements:
+            element = self.get_element(el)
+            mappings = element.exact_mappings + element.close_mappings + element.narrow_mappings + element.broad_mappings
+            if mapping_id in mappings:
+                model_elements.append(element.name)
+        return model_elements
+
     def get_mapping_index(self, imports=True, expand=False) -> Dict[URIorCURIE, List[Tuple[MAPPING_TYPE, Element]]]:
         """
         Returns an index of all elements keyed by the mapping value.
@@ -701,12 +798,11 @@ class SchemaView(object):
         """
         ix = defaultdict(list)
         for en in self.all_elements(imports=imports):
-            for mapping_type, vs in self.get_mappings(en, imports=imports, expand=expand):
+            for mapping_type, vs in self.get_mappings(en, imports=imports, expand=expand).items():
                 for v in vs:
                     ix[v].append((mapping_type, self.get_element(en, imports=imports)))
         return ix
 
-
     @lru_cache()
     def is_relationship(self, class_name: CLASS_NAME = None, imports=True) -> bool:
         """
@@ -788,6 +884,17 @@ class SchemaView(object):
         if slot is not None:
             # case 1: there is an explicit declaration of the slot
             islot = deepcopy(slot)
+            slot_anc_names = self.slot_ancestors(slot_name, reflexive=True)
+            # inheritable slot: first propagate from ancestors
+            for anc_sn in reversed(slot_anc_names):
+                anc_slot = self.get_slot(anc_sn)
+                for metaslot_name in SlotDefinition._inherited_slots:
+                    if getattr(anc_slot, metaslot_name, None):
+                        setattr(islot, metaslot_name, deepcopy(getattr(anc_slot, metaslot_name)))
+            # then override with this
+            #for metaslot_name in SlotDefinition._inherited_slots:
+            #    if getattr(slot, metaslot_name, None):
+            #        setattr(islot, metaslot_name, deepcopy(getattr(slot, metaslot_name)))
         else:
             # case 2: no independent slot definition, declared as an attribute
             # (can be attribute of the class, or an ancestor)
@@ -852,6 +959,26 @@ class SchemaView(object):
         """
         return [self.induced_slot(sn, class_name, imports=imports) for sn in self.class_slots(class_name)]
 
+    @lru_cache()
+    def induced_class(self, class_name: CLASS_NAME = None) -> ClassDefinition:
+        """
+        Generate an induced class
+
+        - the class will have no slots
+        - the class will have one attribute per `class_induced_slots`
+
+        Induced slots are represented as attributes as these are fully locally owner by the class
+        :param class_name: base class name
+        :param imports:
+        :return: induced class
+        """
+        c = deepcopy(self.get_class(class_name))
+        attrs = self.class_induced_slots(c.name)
+        for a in attrs:
+            c.attributes[a.name] = a
+        c.slots = []
+        return c
+
     @lru_cache()
     def get_identifier_slot(self, cn: CLASS_NAME, imports=True) -> Optional[SlotDefinition]:
         """
@@ -865,6 +992,19 @@ class SchemaView(object):
                 return s
         return None
 
+    @lru_cache()
+    def get_type_designator_slot(self, cn: CLASS_NAME, imports=True) -> Optional[SlotDefinition]:
+        """
+        :param cn: class name
+        :param imports:
+        :return: name of slot that acts as type designator for the given class
+        """
+        for sn in self.class_slots(cn, imports=imports):
+            s = self.induced_slot(sn, cn, imports=imports)
+            if s.designates_type:
+                return s
+        return None
+
     def is_inlined(self, slot: SlotDefinition, imports=True) -> bool:
         """
         True if slot is inferred or asserted inline
@@ -890,14 +1030,63 @@ class SchemaView(object):
             else:
                 return False
 
+    def slot_applicable_range_elements(self, slot: SlotDefinition) -> List[ClassDefinitionName]:
+        """
+        Returns all applicable metamodel elements for a slot range
+        (metamodel class names returned: class_definition, enum_definition, type_definition)
+
+        Typically any given slot has exactly one range, and one metamodel element type,
+        but a proposed feature in LinkML 1.2 is range expressions, where ranges can be defined as unions
+
+        Additionally, if linkml:Any is a class_uri then this maps to the any element
+
+        :param slot:
+        :return: list of element types
+        """
+        is_any = False
+        range_types = []
+        for r in self.slot_range_as_union(slot):
+            if r in self.all_classes():
+                range_types.append(ClassDefinition.class_name)
+                rc = self.get_class(r)
+                if rc.class_uri == 'linkml:Any':
+                    is_any = True
+            if is_any or r in self.all_enums():
+                range_types.append(EnumDefinition.class_name)
+            if is_any or r in self.all_types():
+                range_types.append(TypeDefinition.class_name)
+        if not range_types:
+            raise ValueError(f'Unrecognized range: {r}')
+        return range_types
+
+    def slot_range_as_union(self, slot: SlotDefinition) -> List[EnumDefinitionName]:
+        """
+        Returns all applicable ranges for a slot
+
+        Typically any given slot has exactly one range, and one metamodel element type,
+        but a proposed feature in LinkML 1.2 is range expressions, where ranges can be defined as unions
+
+        :param slot:
+        :return: list of ranges
+        """
+        r = slot.range
+        range_union_of = [r]
+        for x in slot.exactly_one_of + slot.any_of:
+            if x.range:
+                range_union_of.append(x.range)
+        return range_union_of
+
     @lru_cache()
     def usage_index(self) -> Dict[ElementName, List[SchemaUsage]]:
         """
-        :return: dictionary keyed by used elements
+        Fetch an index that shows the ways in which each element is used
+
+        :return: dictionary of SchemaUsages keyed by used elements
         """
         ROLES = ['domain', 'range']
         ix = defaultdict(list)
         for cn, c in self.all_classes().items():
+            direct_slots = c.slots
             for sn in self.class_slots(cn):
                 s = self.induced_slot(sn, cn)
                 for k in ROLES:
@@ -907,8 +1096,10 @@ class SchemaView(object):
                     else:
                         vl = [v]
                     for x in vl:
-                        u = SchemaUsage(used_by=cn, slot=sn, metaslot=k, used=x)
-                        ix[x].append(u)
+                        if x is not None:
+                            u = SchemaUsage(used_by=cn, slot=sn, metaslot=k, used=x)
+                            u.inferred = sn in direct_slots
+                            ix[x].append(u)
         return ix
 
     # MUTATION OPERATIONS
